/*
 * arch/arm64/kernel/topology.c
 *
 * Copyright (C) 2011,2013,2014 Linaro Limited.
 *
 * Based on the arm32 version written by Vincent Guittot in turn based on
 * arch/sh/kernel/topology.c
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 */

#include <linux/cpu.h>
#include <linux/cpumask.h>
#include <linux/init.h>
#include <linux/percpu.h>
#include <linux/node.h>
#include <linux/nodemask.h>
#include <linux/of.h>
#include <linux/sched.h>
#include <linux/cpufreq.h>

#include <asm/cputype.h>
#include <asm/topology.h>

static int __init get_cpu_for_node(struct device_node *node)
{
	struct device_node *cpu_node;
	int cpu;

	cpu_node = of_parse_phandle(node, "cpu", 0);
	if (!cpu_node)
		return -1;

	for_each_possible_cpu(cpu) {
		if (of_get_cpu_node(cpu, NULL) == cpu_node) {
			of_node_put(cpu_node);
			return cpu;
		}
	}

	pr_crit("Unable to find CPU node for %s\n", cpu_node->full_name);

	of_node_put(cpu_node);
	return -1;
}

static int __init parse_core(struct device_node *core, int cluster_id,
			     int core_id)
{
	char name[10];
	bool leaf = true;
	int i = 0;
	int cpu;
	struct device_node *t;

	do {
		snprintf(name, sizeof(name), "thread%d", i);
		t = of_get_child_by_name(core, name);
		if (t) {
			leaf = false;
			cpu = get_cpu_for_node(t);
			if (cpu >= 0) {
				cpu_topology[cpu].cluster_id = cluster_id;
				cpu_topology[cpu].core_id = core_id;
				cpu_topology[cpu].thread_id = i;
			} else {
				pr_err("%s: Can't get CPU for thread\n",
				       t->full_name);
				of_node_put(t);
				return -EINVAL;
			}
			of_node_put(t);
		}
		i++;
	} while (t);

	cpu = get_cpu_for_node(core);
	if (cpu >= 0) {
		if (!leaf) {
			pr_err("%s: Core has both threads and CPU\n",
			       core->full_name);
			return -EINVAL;
		}

		cpu_topology[cpu].cluster_id = cluster_id;
		cpu_topology[cpu].core_id = core_id;
	} else if (leaf) {
		pr_err("%s: Can't get CPU for leaf core\n", core->full_name);
		return -EINVAL;
	}

	return 0;
}

static int __init parse_cluster(struct device_node *cluster, int depth)
{
	char name[10];
	bool leaf = true;
	bool has_cores = false;
	struct device_node *c;
	static int cluster_id __initdata;
	int core_id = 0;
	int i, ret;

	/*
	 * First check for child clusters; we currently ignore any
	 * information about the nesting of clusters and present the
	 * scheduler with a flat list of them.
	 */
	i = 0;
	do {
		snprintf(name, sizeof(name), "cluster%d", i);
		c = of_get_child_by_name(cluster, name);
		if (c) {
			leaf = false;
			ret = parse_cluster(c, depth + 1);
			of_node_put(c);
			if (ret != 0)
				return ret;
		}
		i++;
	} while (c);

	/* Now check for cores */
	i = 0;
	do {
		snprintf(name, sizeof(name), "core%d", i);
		c = of_get_child_by_name(cluster, name);
		if (c) {
			has_cores = true;

			if (depth == 0) {
				pr_err("%s: cpu-map children should be clusters\n",
				       c->full_name);
				of_node_put(c);
				return -EINVAL;
			}

			if (leaf) {
				ret = parse_core(c, cluster_id, core_id++);
			} else {
				pr_err("%s: Non-leaf cluster with core %s\n",
				       cluster->full_name, name);
				ret = -EINVAL;
			}

			of_node_put(c);
			if (ret != 0)
				return ret;
		}
		i++;
	} while (c);

	if (leaf && !has_cores)
		pr_warn("%s: empty cluster\n", cluster->full_name);

	if (leaf)
		cluster_id++;

	return 0;
}

static DEFINE_PER_CPU(unsigned long, cpu_capacity);
static DEFINE_PER_CPU(unsigned long, cpu_ipc);

unsigned long arm64_arch_scale_cpu_capacity(struct sched_domain *sd, int cpu)
{
	unsigned long cap = per_cpu(cpu_capacity, cpu);

	if (!cap || cap > SCHED_CAPACITY_SCALE)
		return SCHED_CAPACITY_SCALE;

	return cap;
}

static void update_cpu_ipc(int cpu, unsigned long ipc)
{
	if (ipc > SCHED_CAPACITY_SCALE)
		ipc = SCHED_CAPACITY_SCALE;

	per_cpu(cpu_ipc, cpu) = ipc;
	per_cpu(cpu_capacity, cpu) = ipc;
}

static void __init init_cpu_capacity(void)
{
	int cpu;
	struct device_node *cn;

	for_each_possible_cpu(cpu) {
		u32 val;

		cn = of_get_cpu_node(cpu, NULL);
		if (!cn) {
			pr_err("%s: CPU%d device node missing\n",
				__func__, cpu);
			return;
		}

		if (of_property_read_u32(cn, "cpu-ipc", &val)) {
			pr_info("%s: Missing cpu-ipc property\n",
				cn->full_name);
			continue;
		}

		pr_info("CPU%d ipc=%u\n", cpu, val);

		update_cpu_ipc(cpu, val);
	}
}

static int __init parse_dt_topology(void)
{
	struct device_node *cn, *map;
	int ret = 0;
	int cpu;

	cn = of_find_node_by_path("/cpus");
	if (!cn) {
		pr_err("No CPU information found in DT\n");
		return 0;
	}

	/*
	 * When topology is provided cpu-map is essentially a root
	 * cluster with restricted subnodes.
	 */
	map = of_get_child_by_name(cn, "cpu-map");
	if (!map)
		goto out;

	ret = parse_cluster(map, 0);
	if (ret != 0)
		goto out_map;

	/*
	 * Check that all cores are in the topology; the SMP code will
	 * only mark cores described in the DT as possible.
	 */
	for_each_possible_cpu(cpu)
		if (cpu_topology[cpu].cluster_id == -1)
			ret = -EINVAL;

out_map:
	of_node_put(map);
out:
	of_node_put(cn);
	return ret;
}

/*
 * cpu topology table
 */
struct cpu_topology cpu_topology[NR_CPUS];
EXPORT_SYMBOL_GPL(cpu_topology);

const struct cpumask *cpu_coregroup_mask(int cpu)
{
	return &cpu_topology[cpu].core_sibling;
}

static void update_siblings_masks(unsigned int cpuid)
{
	struct cpu_topology *cpu_topo, *cpuid_topo = &cpu_topology[cpuid];
	int cpu;

	/* update core and thread sibling masks */
	for_each_possible_cpu(cpu) {
		cpu_topo = &cpu_topology[cpu];

		if (cpuid_topo->cluster_id != cpu_topo->cluster_id)
			continue;

		cpumask_set_cpu(cpuid, &cpu_topo->core_sibling);
		if (cpu != cpuid)
			cpumask_set_cpu(cpu, &cpuid_topo->core_sibling);

		if (cpuid_topo->core_id != cpu_topo->core_id)
			continue;

		cpumask_set_cpu(cpuid, &cpu_topo->thread_sibling);
		if (cpu != cpuid)
			cpumask_set_cpu(cpu, &cpuid_topo->thread_sibling);
	}
}

void store_cpu_topology(unsigned int cpuid)
{
	struct cpu_topology *cpuid_topo = &cpu_topology[cpuid];
	u64 mpidr;

	if (cpuid_topo->cluster_id != -1)
		goto topology_populated;

	mpidr = read_cpuid_mpidr();

	/* Uniprocessor systems can rely on default topology values */
	if (mpidr & MPIDR_UP_BITMASK)
		return;

	/* Create cpu topology mapping based on MPIDR. */
	if (mpidr & MPIDR_MT_BITMASK) {
		/* Multiprocessor system : Multi-threads per core */
		cpuid_topo->thread_id  = MPIDR_AFFINITY_LEVEL(mpidr, 0);
		cpuid_topo->core_id    = MPIDR_AFFINITY_LEVEL(mpidr, 1);
		cpuid_topo->cluster_id = MPIDR_AFFINITY_LEVEL(mpidr, 2) |
					 MPIDR_AFFINITY_LEVEL(mpidr, 3) << 8;
	} else {
		/* Multiprocessor system : Single-thread per core */
		cpuid_topo->thread_id  = -1;
		cpuid_topo->core_id    = MPIDR_AFFINITY_LEVEL(mpidr, 0);
		cpuid_topo->cluster_id = MPIDR_AFFINITY_LEVEL(mpidr, 1) |
					 MPIDR_AFFINITY_LEVEL(mpidr, 2) << 8 |
					 MPIDR_AFFINITY_LEVEL(mpidr, 3) << 16;
	}

	pr_debug("CPU%u: cluster %d core %d thread %d mpidr %#016llx\n",
		 cpuid, cpuid_topo->cluster_id, cpuid_topo->core_id,
		 cpuid_topo->thread_id, mpidr);

topology_populated:
	update_siblings_masks(cpuid);
}

#ifdef CONFIG_CPU_FREQ

static DEFINE_MUTEX(capacity_update_lock);
static DEFINE_PER_CPU(unsigned long, cur_freq);
static DEFINE_PER_CPU(unsigned long, max_freq);
static DEFINE_PER_CPU(unsigned long, freq_cap);
static DEFINE_PER_CPU(unsigned long, raw_cpu_capacity);

unsigned long arm64_arch_scale_freq_capacity(struct sched_domain *sd, int cpu)
{
	unsigned long freq_capacity = per_cpu(freq_cap, cpu);

	if (!freq_capacity || freq_capacity > SCHED_CAPACITY_SCALE)
		return SCHED_CAPACITY_SCALE;

	return freq_capacity;
}

static void update_raw_cpu_capacity(int cpu)
{
	unsigned long ipc = per_cpu(cpu_ipc, cpu);
	unsigned long max = per_cpu(max_freq, cpu);

	if (!ipc || !max)
		return;

	per_cpu(raw_cpu_capacity, cpu) = ipc * max;
}

/*
 * Normalize raw cpu capacities relative to max raw cpu capacity
 * Must be called with capacity_update_lock held.
 */
static void update_cpu_capacities(void)
{
	unsigned long max_raw_capacity = 0;
	int i;

	for_each_online_cpu(i) {
		unsigned long capacity = per_cpu(raw_cpu_capacity, i);

		if (capacity > max_raw_capacity)
			max_raw_capacity = capacity;
	}

	if (!max_raw_capacity)
		return;

	for_each_possible_cpu(i) {
		unsigned long capacity = per_cpu(raw_cpu_capacity, i);

		capacity *= SCHED_CAPACITY_SCALE;
		capacity /= max_raw_capacity;
		if (capacity > SCHED_CAPACITY_SCALE)
			capacity = SCHED_CAPACITY_SCALE;

		per_cpu(cpu_capacity, i) = capacity;
		pr_debug("%s: cpu=%d capacity=%lu\n", __func__, i, capacity);
	}
}

static void update_freq_capacity(int cpu)
{
	unsigned long cur = per_cpu(cur_freq, cpu);
	unsigned long max = per_cpu(max_freq, cpu);

	if (!max || !cur) {
		per_cpu(freq_cap, cpu) = 0;
		return;
	}

	if (cur >= max)
		per_cpu(freq_cap, cpu) = SCHED_CAPACITY_SCALE;
	else
		per_cpu(freq_cap, cpu) = (cur * SCHED_CAPACITY_SCALE) / max;
}

static int cpufreq_callback(struct notifier_block *nb,
					unsigned long val, void *data)
{
	struct cpufreq_freqs *freqs = data;
	int cpu = freqs->cpu;

	if (val != CPUFREQ_POSTCHANGE)
		return NOTIFY_OK;

	per_cpu(cur_freq, cpu) = freqs->new;
	update_freq_capacity(cpu);

	return NOTIFY_OK;
}

static struct notifier_block cpufreq_notifier = {
	.notifier_call  = cpufreq_callback,
};

static int cpufreq_policy_callback(struct notifier_block *nb,
					unsigned long val, void *data)
{
	int i;
	struct cpufreq_policy *policy = data;
	bool need_update = false;

	if ((val != CPUFREQ_NOTIFY && val != CPUFREQ_START) || !policy)
		return NOTIFY_OK;

	mutex_lock(&capacity_update_lock);

	for_each_cpu(i, policy->cpus) {
		if (policy->max == per_cpu(max_freq, i))
			continue;
		per_cpu(max_freq, i) = policy->max;
		update_freq_capacity(i);
		update_raw_cpu_capacity(i);
		need_update = true;
	}

	/* Re-normalize capacities */
	if (need_update)
		update_cpu_capacities();

	mutex_unlock(&capacity_update_lock);

	return NOTIFY_OK;
}

static struct notifier_block cpufreq_policy_notifier = {
	.notifier_call = cpufreq_policy_callback,
};

static int cpu_hotplug_callback(struct notifier_block *nb,
					unsigned long action, void *hcpu)
{
	unsigned int cpu = (unsigned long)hcpu;

	if (action != CPU_DEAD && action != CPU_DEAD_FROZEN)
		return NOTIFY_OK;

	mutex_lock(&capacity_update_lock);
	per_cpu(max_freq, cpu) = 0;
	per_cpu(raw_cpu_capacity, cpu) = 0;
	update_cpu_capacities();
	mutex_unlock(&capacity_update_lock);

	return NOTIFY_OK;
}

static struct notifier_block cpu_hotplug_notifier = {
	.notifier_call = cpu_hotplug_callback,
};

static int __init register_notifiers(void)
{
	int ret;

	ret = cpufreq_register_notifier(&cpufreq_notifier,
						CPUFREQ_TRANSITION_NOTIFIER);
	if (ret)
		return ret;

	ret = cpufreq_register_notifier(&cpufreq_policy_notifier,
						CPUFREQ_POLICY_NOTIFIER);
	if (ret)
		return ret;

	return register_cpu_notifier(&cpu_hotplug_notifier);
}
core_initcall(register_notifiers);
#endif

static void __init reset_cpu_topology(void)
{
	unsigned int cpu;

	for_each_possible_cpu(cpu) {
		struct cpu_topology *cpu_topo = &cpu_topology[cpu];

		cpu_topo->thread_id = -1;
		cpu_topo->core_id = 0;
		cpu_topo->cluster_id = -1;

		cpumask_clear(&cpu_topo->core_sibling);
		cpumask_set_cpu(cpu, &cpu_topo->core_sibling);
		cpumask_clear(&cpu_topo->thread_sibling);
		cpumask_set_cpu(cpu, &cpu_topo->thread_sibling);
	}
}

void __init init_cpu_topology(void)
{
	reset_cpu_topology();

	/*
	 * Discard anything that was parsed if we hit an error so we
	 * don't use partial information.
	 */
	if (of_have_populated_dt() && parse_dt_topology())
		reset_cpu_topology();

	init_cpu_capacity();
}
