/*
 * tegra_virt_ref_alt.c - Tegra reference virtual machine driver
 *
 * Copyright (c) 2015-2017, NVIDIA CORPORATION.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/module.h>
#include <linux/of.h>
#include <linux/platform_device.h>
#include <sound/soc.h>
#include <linux/of_platform.h>
#include <linux/pm_runtime.h>
#include <linux/tegra_pm_domains.h>

#include "tegra_dtcp_sad.h"
#include "tegra210_virt_alt_admaif.h"
#include "tegra_asoc_machine_virt_alt.h"
#include "tegra_asoc_util_virt_alt.h"

static const struct snd_soc_dapm_widget tegra_virt_dapm_widgets[] = {
	SND_SOC_DAPM_HP("Headphone", NULL),
	SND_SOC_DAPM_LINE("LineIn", NULL),
};

static const struct snd_soc_dapm_route tegra_virt_dapm_routes[] = {
	{"Headphone", NULL, "OUT"},
	{"IN", NULL, "LineIn"},
};


static sad_context_t sad = {
	.sad_mode = SAD_HEADER_MODE,
	.init_sad_flood = 0,
	.enable_sad_flood = 0,

	/* 0 indexed */
	.admaif_id = 3,
	.dma_id = 3,
	.dma_ch_page = 0,
};

static struct snd_soc_card tegra_virt_t210ref_card = {
	.name = "t210ref-virt-card",
	.owner = THIS_MODULE,
	.dapm_widgets = tegra_virt_dapm_widgets,
	.num_dapm_widgets = ARRAY_SIZE(tegra_virt_dapm_widgets),
	.dapm_routes = tegra_virt_dapm_routes,
	.num_dapm_routes = ARRAY_SIZE(tegra_virt_dapm_routes),
	.fully_routed = true,
};

static struct snd_soc_card tegra_virt_t186ref_card = {
	.name = "t186ref-virt-card",
	.owner = THIS_MODULE,
	.dapm_widgets = tegra_virt_dapm_widgets,
	.num_dapm_widgets = ARRAY_SIZE(tegra_virt_dapm_widgets),
	.dapm_routes = tegra_virt_dapm_routes,
	.num_dapm_routes = ARRAY_SIZE(tegra_virt_dapm_routes),
	.fully_routed = true,
};

static void tegra_virt_set_dai_params(
		struct snd_soc_dai_link *dai_link,
		struct snd_soc_pcm_stream *user_params,
		unsigned int dai_id)
{
	dai_link[dai_id].params = user_params;
}

static struct tegra_virt_admaif_soc_data soc_data_tegra210 = {
	.num_ch = TEGRA210_ADMAIF_CHANNEL_COUNT,
};

static struct tegra_virt_admaif_soc_data soc_data_tegra186 = {
	.num_ch = TEGRA186_ADMAIF_CHANNEL_COUNT,
};


static const struct of_device_id tegra_virt_machine_of_match[] = {
	{ .compatible = "nvidia,tegra210-virt-pcm",
		.data = &soc_data_tegra210 },
	{ .compatible = "nvidia,tegra186-virt-pcm",
		.data = &soc_data_tegra186},
	{},
};

static int tegra_virt_machine_driver_probe(struct platform_device *pdev)
{
	struct snd_soc_card *card = NULL;
	int i, ret = 0;
	int admaif_ch_num = 0;
	bool adsp_enabled = false;
	unsigned int admaif_ch_list[MAX_ADMAIF_IDS];
	const struct of_device_id *match;
	struct tegra_virt_admaif_soc_data *soc_data;

	match = tegra_virt_machine_of_match;
	if (of_device_is_compatible(pdev->dev.of_node,
		"nvidia,tegra210-virt-pcm")) {
		card = &tegra_virt_t210ref_card;
		match = of_match_device(tegra_virt_machine_of_match,
						&pdev->dev);
		soc_data = (struct tegra_virt_admaif_soc_data *)match->data;
	} else if (of_device_is_compatible(pdev->dev.of_node,
		"nvidia,tegra186-virt-pcm")) {
		card = &tegra_virt_t186ref_card;
		match = of_match_device(tegra_virt_machine_of_match,
						&pdev->dev);
		soc_data = (struct tegra_virt_admaif_soc_data *)match->data;
	} else {
		dev_err(&pdev->dev, "device match data not found\n");
		return -ENODEV;
	}

	card->dev = &pdev->dev;
	card->dai_link = tegra_virt_machine_get_dai_link();
	card->num_links = tegra_virt_machine_get_num_dai_links();
	adsp_enabled = of_property_read_bool(pdev->dev.of_node,
		"adsp_enabled");


	if (adsp_enabled) {
		dev_info(&pdev->dev, "virt-alt-pcm: adsp config is set\n");
	} else {
		dev_info(&pdev->dev, "virt-alt-pcm: adsp config is not set\n");
		card->num_links = soc_data->num_ch;
	}

	for (i = 0; i < card->num_links; i++) {
		/* each ADMAIF has one corresponding ADSP-ADMAIF node,
		* thus doing multiplication by 2
		*/
		if (i < (soc_data->num_ch * 2)) {
			card->dai_link[i].codec_of_node =
			of_parse_phandle(pdev->dev.of_node, "codec", 0);
		}
		if (i < soc_data->num_ch) {
			strcpy((char *)card->dai_link[i].cpu_name,
					pdev->name);
		}
	}

	if (tegra210_virt_admaif_register_component(pdev, soc_data)) {
		dev_err(&pdev->dev, "Failed register admaif component\n");
		return -EINVAL;
	}

	if (of_property_read_u32(pdev->dev.of_node,
				"admaif_ch_num", &admaif_ch_num)) {
		dev_err(&pdev->dev, "number of admaif channels is not set\n");
		return -EINVAL;
	}

	if (of_property_read_string(pdev->dev.of_node, "cardname", &card->name))
		dev_warn(&pdev->dev, "Using default card name %s\n",
			card->name);

	if (admaif_ch_num > 0) {

		if (of_property_read_u32_array(pdev->dev.of_node,
						"admaif_ch_list",
						admaif_ch_list,
						admaif_ch_num)) {
			dev_err(&pdev->dev, "admaif_ch_list os not populated\n");
			return -EINVAL;
		}
		for (i = 0; i < admaif_ch_num; i++) {
			tegra_virt_set_dai_params(
					tegra_virt_machine_get_dai_link(),
					NULL,
					(admaif_ch_list[i] - 1));
		}
	}

	ret = snd_soc_register_card(card);
	if (ret) {
		dev_err(&pdev->dev, "snd_soc_register_card failed (%d)\n",
			ret);
		return -EINVAL;
	}
	tegra_setup_sad(pdev, &sad, card);
	tegra_pd_add_device(&pdev->dev);
	pm_runtime_forbid(&pdev->dev);

	return ret;
}

static int tegra_virt_machine_driver_remove(struct platform_device *pdev)
{
	struct snd_soc_card *card = platform_get_drvdata(pdev);

	snd_soc_unregister_card(card);

	return 0;
}

static struct platform_driver tegra_virt_machine_driver = {
	.driver = {
		.name = DRV_NAME,
		.owner = THIS_MODULE,
		.pm = &snd_soc_pm_ops,
		.of_match_table =
			of_match_ptr(tegra_virt_machine_of_match),
	},
	.probe = tegra_virt_machine_driver_probe,
	.remove = tegra_virt_machine_driver_remove,
};
module_platform_driver(tegra_virt_machine_driver);

MODULE_AUTHOR("Paresh Anandathirtha <paresha@nvidia.com>");
MODULE_DESCRIPTION("Tegra virt machine driver");
MODULE_LICENSE("GPL");
MODULE_DEVICE_TABLE(of, tegra_virt_machine_of_match);
MODULE_ALIAS("platform:" DRV_NAME);
